/*
  CPP_CONTEST=2013Cal-2012
  CPP_PROBLEM=R
  CPP_LANG=C+OPENMP+MPI
  CPP_PROCESSES_PER_NODE=marte 1 mercurio 1
*/


/* RECORD
María Asunción Sánchez López
student of Methodology of  Parallel Programming,
at the University of Murcia, Spain
January 27, 2013
time 2038 msec
speed-up 7.9
*/ 

/*RECORD in the new system
when the programs in the record table are run in the new system
formed by marte+mercurio (two hexacores),
it is the absolute record, with
time 3731 msec
speed-up 4.37
*/

#include <stdlib.h>
#include <stdio.h>
#include "mpi.h"

#define MARK 5 //this value in a position in the checkerboard indicates the position has been marked

#ifdef DEBUG
  extern void viewcheckerboard(int,int *);
#endif

int impulsing(int n,int *cb,int ix,int iy,int *fx,int *fy)
{
  int temp;

//if the position is out of the checkerboard do nothing
  if(ix<0 || ix>=n || iy<0 || iy>=n)
  {
    return 0;
  }
  temp=cb[ix*n+iy];
//if the position is marked or an impulsing position do nothing
  if(1<=temp && temp<=MARK)
  {
    return 0;
  }
//if the position is empty, store it and indicate a position has been found
  if(temp==0)
  {
    *fx=ix;
    *fy=iy;
    cb[ix*n+iy]=MARK;
    return 1;
  }
  return 0;
}

int move(int n,int *cb,int ix,int iy,int *fx,int *fy)
{
//if the position is out of the checkerboard do nothing
  if(ix<0 || ix>=n || iy<0 || iy>=n)
  {
    return 0;
  }
//if the position is marked do nothing
  if(cb[ix*n+iy]==MARK)
  {
    return 0;
  }
//if the position is empty, store it and indicate a position has been found
  if(cb[ix*n+iy]==0)
  {
    *fx=ix;
    *fy=iy;
    cb[ix*n+iy]=MARK;
    return 1;
  }
//impulsing to the left
  if(cb[ix*n+iy]==1)
  {
    return(impulsing(n,cb,ix,iy-1,fx,fy));
  }
  else if(cb[ix*n+iy]==2)
  {
    return(impulsing(n,cb,ix-1,iy,fx,fy));
  }
  else if(cb[ix*n+iy]==3)
  {
    return(impulsing(n,cb,ix,iy+1,fx,fy));
  }
  else if(cb[ix*n+iy]==4)
  {
    return(impulsing(n,cb,ix+1,iy,fx,fy));
  }
  return 0;
}

int minimummov(int n,int *cb,int sx,int sy,int tx,int ty)
{
  int *activesx,*activesy; //coordenates of the actives positions
  int nactives; //number of active positions
  int *tempx,*tempy; //auxiliar vector to store temporally the coordenates of the actives positions
  int *temp,i,j;
  int step,end,foundposition;
  char c;

  if(sx==tx && sy==ty)
    return 0;
    
  activesx=(int *) malloc(sizeof(int)*n*n);
  activesy=(int *) malloc(sizeof(int)*n*n);
  tempx=(int *) malloc(sizeof(int)*n*n);
  tempy=(int *) malloc(sizeof(int)*n*n);

  cb[sx*n+sy]=MARK; //the source is marked
 //the source is marked

  activesx[0]=sx;
  activesy[0]=sy;
  nactives=1;
  step=1;
  end=0;
//  viewcheckerboard(n,cb);

  do
  {
     i=0;
     j=0;
     while(i<nactives && end==0)
    {
       foundposition=move(n,cb,activesx[i],activesy[i]-1,&tempx[j],&tempy[j]); //moves to the left. If movement is posible the position is stored and the number of stored positions is incremented
#ifdef DEBUG
  printf("Step %d. After moving to the left, from %d,%d, found %d in position %d,%d:\n",step,activesx[i],activesy[i],foundposition,tempx[j],tempy[j]);
  viewcheckerboard(n,cb);
  scanf("%c\n",&c);
#endif
       if(foundposition)
       {
          if(tempx[j]==tx && tempy[j]==ty) //the stored position is the target position
          {
             end=1;
          }
          j++;
       }
      if(end==0)
       {
         foundposition=move(n,cb,activesx[i]-1,activesy[i],&tempx[j],&tempy[j]);
#ifdef DEBUG
  printf("Step %d. After moving up, from %d,%d, found %d in position %d,%d:\n",step,activesx[i],activesy[i],foundposition,tempx[j],tempy[j]);
  viewcheckerboard(n,cb);
  scanf("%c\n",&c);
#endif
         if(foundposition)
         {
            if(tempx[j]==tx && tempy[j]==ty) //the stored position is the target position
            { 
               end=1;
            }
            j++;
         }
       }
       if(end==0)
       {
         foundposition=move(n,cb,activesx[i],activesy[i]+1,&tempx[j],&tempy[j]);
#ifdef DEBUG
  printf("Step %d. After moving to the right, from %d,%d, found %d in position %d,%d:\n",step,activesx[i],activesy[i],foundposition,tempx[j],tempy[j]);
  viewcheckerboard(n,cb);
  scanf("%c\n",&c);
#endif
         if(foundposition)
         {
            if(tempx[j]==tx && tempy[j]==ty) //the stored position is the target position
            {
               end=1;
            }
            j++;
         }
       }
       if (end==0)
       {
         foundposition=move(n,cb,activesx[i]+1,activesy[i],&tempx[j],&tempy[j]);
#ifdef DEBUG
  printf("Step %d. After moving down, from %d,%d, found %d in position %d,%d:\n",step,activesx[i],activesy[i],foundposition,tempx[j],tempy[j]);
  viewcheckerboard(n,cb);
  scanf("%c\n",&c);
#endif
         if(foundposition)
         {
            if(tempx[j]==tx && tempy[j]==ty) //the stored position is the target position
            {
               end=1;
            }
            j++;
         }
       }
      i++;
    } //while
    nactives=j;
    temp=activesx;
    activesx=tempx;
    tempx=temp;
    temp=activesy;
    activesy=tempy;
    tempy=temp;
    step++;
  } while(nactives && !end);
  step--;
  free(activesx);
  free(activesy);
  free(tempx);
  free(tempy);
#ifdef DEBUG
  printf("Minimo de %d,%d a %d,%d: %d\n\n",sx,sy,tx,ty,step);
#endif
  //viewcheckerboard(n,cb);
  return step;
}

void sec(int n,int *cb,int p,int *positions,int *movements,int node,int np)
{
	int i,j, inicio; //auxiliar
	int n_parejas = p/np;
	int resto = p%np;
	MPI_Status status;

	if (node < resto) {
		n_parejas = n_parejas + 1;
		inicio = (node * n_parejas)*4;
	}
	else inicio = (node*n_parejas + resto)*4;

	if (node != 0) {
		cb = (int *) malloc(sizeof(int)*n*n);
		positions = (int *) malloc(sizeof(int)*p*4);
		movements=(int *) malloc(sizeof(int)*n_parejas);
	}

	MPI_Bcast(cb,n*n,MPI_INT,0,MPI_COMM_WORLD);
	MPI_Bcast(positions,p*4,MPI_INT,0,MPI_COMM_WORLD);
        omp_set_num_threads(6);
	#pragma omp parallel private(i,j)
	{
	int *copycb=(int *) malloc(sizeof(int)*n*n);

	for(i=0;i<n*n;i++) copycb[i]=cb[i];

	#pragma omp for schedule(dynamic)
	for(i=0;i<n_parejas;i++)
	{
		movements[i]=minimummov(n,copycb,positions[inicio + i*4],positions[inicio + i*4+1],positions[inicio + i*4+2],positions[inicio + i*4+3]);
		for(j=0;j<n*n;j++) copycb[j]=cb[j];
	}
	free(copycb);
	}
	if (node == 0){
		int posicion = n_parejas;
		for (i = 1; i< np; i++){
			if (i < resto || resto == 0){
				MPI_Recv(&movements[posicion],n_parejas, MPI_INT, i, 5, MPI_COMM_WORLD, &status);
				posicion = posicion + n_parejas;
			}
			else {	
				if ((n_parejas-1)!= 0 ){			
					MPI_Recv(&movements[posicion],n_parejas-1, MPI_INT, i, 5, MPI_COMM_WORLD, &status);
					posicion = posicion + n_parejas - 1;
				}
			}
		}
	}
	else {
		if (n_parejas > 0){

			MPI_Send(&movements[0],n_parejas, MPI_INT, 0, 5, MPI_COMM_WORLD);
		}
		free(cb);
		free(movements);
		free(positions);
	}

}
